
package Caraca;

require 5.000;

use IO::Socket::UNIX;
use Carp;
use strict;
use vars qw(@ISA $VERSION);
use Exporter;

# legacy

$VERSION = "0.25";

sub import {
    my $pkg = shift;
    my $callpkg = caller;
    Exporter::export 'Caraca', $callpkg, @_;
}

sub new {
    my($class,%arg) = @_;
    my $self = {};
    $self->{SOCK} = IO::Socket::UNIX->new(Local => "/var/tmp/$$",
		    Peer  => '/tmp/caracad',
		    Type => SOCK_STREAM);

    #$sock->autoflush(1);
    bless ($self, $class);
    return $self;
}

sub send {
	my $self = shift;
	my $answer='';
	$self->{SOCK}->send(shift);
	return $self->{SOCK}->recv($answer,100,0);
	return 0 if $answer =~ m/^OK/;
	warn $answer;
}
sub rtr {
	my $self = $_[0];
	my $answer='';
	$self->send('RTR '.$_[1]." 0\0");
	$self->{SOCK}->recv($answer,100,0);
	$answer =~ s/.*://;
	return $answer;
}

sub filter {
	my $self = $_[0];
	
	$self->send('filter '.$_[1].' '.$_[2]." 0\0");
	
}

1;

__END__

=head1 NAME

Caraca - Object interface to CAN communication via caracad

=head1 SYNOPSIS

    use Caraca;

=head1 DESCRIPTION

C<Caraca> provides an object interface to communicate over the CAN-Bus. It
is built upon the L<IO::Socket> interface.

=head1 CONSTRUCTOR

=over 4

=item new ( [ARGS] )

Creates an connection via unix sockets to caracad (our message dispatcher for
CAN-Messages.

=back

=head1 METHODS

=over 4

=item filter(id,mask)

We set an filter to receive Messages from the CAN-Bus. If 

=item rtr(id)

Sends an Message with the given id and RTR-Bit set. So we can query our sensors.


=back

=head1 SEE ALSO

L<IO::Socket::UNIX>

=head1 AUTHOR

Konrad Riedel E<lt>F<k.riedel@gmx.de>E<gt>

=head1 COPYRIGHT

Copyright (c) 2000 Konrad Riedel <k.riedel@gmx.de>. All rights reserved.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=cut
