#include   <termios.h>
#include   <unistd.h>
#include   <fcntl.h>
#include   <stdio.h>
#include   "can_proto.h"

static struct termios   alt_terminal;
static enum { RESET, RAW, CBREAK } tty_modus = RESET;

int tty_set(int fd)
{
   struct termios  terminal;

   if (tcgetattr(fd, &alt_terminal) < 0)
      return(-1);

   terminal = alt_terminal;
   
   if (cfsetospeed(&terminal,B38400) < 0)
      return(-1);
   if (cfsetispeed(&terminal,B38400) < 0)
      return(-1);
  
    /* ECHO, kanonischen Modus, erweitert. Zeichensatz und
       Signal-Zeichen ausschalten */
   terminal.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG); 

    /* kein SIGINT bei BREAK, kein Umwandeln von CR nach NL,
       keine Eingabe-Paritaetspruefung, kein Abschneiden des 8.Bits,
       und  kein START/STOP-Ausgabeprotokoll */
   terminal.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON); 

    /* alte Zeichengrosse und Paritaetspruefung ausschalten */
   terminal.c_cflag &= ~(CSIZE | PARENB); 
   terminal.c_cflag |= CS8; /* 8 Bits pro Zeichen setzen */

    /* Spezielle implementations-definierte Ausgabeart ausschalten */
   terminal.c_oflag &= ~(OPOST); 

    /* Fall 2: Immer nur 1 Byte; kein Timer */
   terminal.c_cc[VMIN] = 1;
   terminal.c_cc[VTIME] = 0;
  
   if (tcsetattr(fd, TCSAFLUSH, &terminal) < 0)
      return(-1);
   return(0);
}

/*------ tty_reset --- Terminal in alten Modus zuruecksetzen ---------*/
int tty_reset(int fd)
{
   if (tty_modus != CBREAK && tty_modus != RAW)
      return(0);
   if (tcsetattr(fd, TCSAFLUSH, &alt_terminal) < 0)
      return(-1);

   tty_modus = RAW;
   return(0);
}



int can_Init(char *device,int flags)
{
	int fd;
	if (( fd = open(device, flags|O_SYNC)) > 0 ) {
		tty_set(fd);
		return(fd);
	}
	return(-1);
}

int can_Receive(int fd,canmsg_t *rx )
{          
	/* we need min == 0 and time == 0 */	
	/* read Header 2Bytes */
	unsigned char id1,id2,start;
	struct timezone  tz;
	int i=0;

	while (read(fd, &start, 1) == 1) {
		//if (start == 'C') printf("\nC>"); else printf("%x,",(int)start);
		if (start == 'C') break; else {
                      printf("%0x,",(int)start);
			if (i++ >20) {printf("\n");i=0;}
		}
	}

	//return(1); 
	gettimeofday(&(rx->timestamp), &tz);
	read(fd, &id1, 1);
	read(fd, &id2, 1);
	rx->id = (id1 << 3) + (id2 >> 5);
	rx->length = (id2 & 0x0f);
	/* read Message */
	read(fd, rx->data, rx->length);
	rx->data[rx->length] = 0;
	rx->flags = 0;
	return(1);   
}
/* must be atomic!*/
int can_Send(int fd,canmsg_t *tx)
{
	//extern int errno;
	unsigned char rtr=0;
	char buf[20];
	char *ptr;
	ptr = buf;
	*ptr++ = 'C';
	*ptr++ = tx->id >> 3;
	if (tx->flags & MSG_RTR) {
		tx->length=0;
		rtr = 0x10;
	}
	
	*ptr++ = ((tx->id & 7) << 5) + (tx->length & 0x0f);
	memcpy(ptr,tx->data,tx->length);
        ptr+=tx->length;
	*ptr++ = 0x0ff;
	return(write(fd, buf, ptr - buf));
}


