#include <syslog.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include "caracad.h"
#include "globals.h"

char     err_msg[MAX_ZEICHEN];


/*----- Prototypen fuer lokale Funktionen -----------------------------*/
extern Globals	*init_globals();
static void  client_alloc(void);
static int   client_add(int fd, pid_t pid);
static void  client_del(Client *client);
inline int serve(void);
int addrtoi(char *addr);
int can_fmt_msg(char *res, canmsg_t *rx );
int send_can2client(canmsg_t *rx,int fd,int format);
static void  request(char *buf, int byte_gelesen, Client *client);
static int   buf_argv(char *buf, int *argc, char *argv[]);
int  serv_init(const char *name);
int  serv_accept(int initfd, uid_t *uidzgr);
int  send_fd(int spipefd , int fd);
int send_ok(int spipefd, const char *errmsg);
int send_err(int spipefd, int status, const char *errmsg);

void usage( char *name )
{
    fprintf( stderr, "usage: %s\t[-d debug-level] [-s syslog-level]\n", name );
    fprintf( stderr, "\t\t[-f CAN device]\n" );
}



/*
 * Interpret argv, and fill in any appropriate globals.
 */
int doargs(int argc, char **argv, Globals *g)
{
	int		i;
	extern char	*optarg;
	int		rflag = 0;

	if ( (g->myname = strrchr( argv[0], '/' )) == NULL ) {
		g->myname = strdup( argv[0] );
	} else {
		g->myname = strdup( g->myname + 1 );
	}

	while ( (i = getopt( argc, argv, "hd:f:r:t:k:o" )) != EOF ) {
		switch ( i ) {
		case 'd':	/* turn on debugging */
			if ( optarg[0] == '?' ) {
				printf( "Debug levels:\n" );
				printf( "\tDEBUG_TRACE\t%d\n", DEBUG_TRACE );
				printf( "\tDEBUG_MSGS\t%d\n", DEBUG_MSGS );
				printf( "\tDEBUG_ARGS\t%d\n", DEBUG_ARGS );
				printf( "\tDEBUG_CONNS\t%d\n", DEBUG_CONNS );
				printf( "\tDEBUG_FILTER\t%d\n", DEBUG_FILTER );
				printf( "\tDEBUG_CONFIG\t%d\n", DEBUG_CONFIG );
				printf( "\tDEBUG_ANY\t%d\n", DEBUG_ANY );
				return( -1 );
			} else {
				debug = atoi( optarg );
			}
			break;
		case 'f':	/* can device */
			g->can_device= strdup( optarg );
			break;
		case 'h':
			usage( g->myname );
			return( -1 );
		default:
			usage( g->myname );
			return( -1 );
		}
	}

	openlog( g->myname, OPENLOG_OPTIONS, LOG_DAEMON );
	return 0;

}


int main(int argc, char *argv[])
{
	/* 
	 * Create and initialize globals.  init_globals() also initializes
	 * the CAN-connection.
	 */
	if (( sglob = init_globals()) == NULL ) {
		fprintf( stderr, "Out of memory initializing globals\n" );
		exit( 1 );
	}
	/*
	 * Read config file and initialize  structs.
	 */
	/*
	 * Process command-line args and fill in globals.
	 */
	if ( doargs( argc, argv, sglob ) < 0 ) {
		exit( 1 );
	}

	printf("using CAN device %s\n", sglob->can_device);

	if (( sglob->can_fd = can_Init(sglob->can_device, O_RDWR)) < 0 ) {
		fprintf(stderr,"Error opening CAN device %s\n", sglob->can_device);
		exit(1);
	}

	//if (debug == 0) daemon(0,0);

	return(serve());
}

static void client_alloc(void)
{
	int   i;

	sglob->client = realloc(sglob->client, (sglob->client_cnt+REALLOC_CNT) * sizeof(Client));
	if (sglob->client == NULL)
		syslog(LOG_ERR, "out of memory (client_alloc)");

	for (i=sglob->client_cnt; i<sglob->client_cnt+REALLOC_CNT; i++)
		sglob->client[i].fd = -1;
	sglob->client_cnt += REALLOC_CNT;
}

static int  client_add(int fd, pid_t pid)
{
	int   i;

	do {
		for (i=0; i<sglob->client_cnt; i++) {
			if (sglob->client[i].fd == -1) {
				sglob->client[i].fd = fd;
				sglob->client[i].pid = pid;
				return(i);
			}
		}
		client_alloc(); /* notwendig, da kein freier Platz gefunden */
	} while (1);
}
/* adding new filter after existing filter f
 */
static struct can_filter *filter_create(struct can_filter *f)
{
	 struct can_filter *fnew = malloc(sizeof(struct can_filter));
	 if (f == NULL)                /* first one */
		 sglob->first_can_filter = sglob->last_can_filter = fnew;
	 else {
		 fnew->next = f->next;
	 }
	 fnew->counter = 0;
	 return fnew;
}

static void filter_del(struct can_filter *f)
{
	struct can_filter *cur, *prev;

	for (cur=prev=sglob->first_can_filter; cur; prev=cur, cur=cur->next) 
		if (cur == f) {
			prev->next = cur->next;
			syslog(LOG_DEBUG, "drop filter for pid %d", cur->client->pid);
			free(cur);
			break;
		}
}

static void client_del(Client *client)
{
	int   i;
	struct can_filter *f;
	for (f = sglob->first_can_filter; f; f = f->next)
		if (f->client == client) 
			filter_del(f);

	for (i=0; i<sglob->client_cnt; i++) {
		if (&sglob->client[i] == client) {
			sglob->client[i].fd = -1;
			return;
		}
	}
	syslog(LOG_ERR, "kein Client-Eintrag zu Filedeskr. %d gefunden", client[i].fd);
}

static void can_dispatch(canmsg_t *rx)
{
	struct can_filter *i;
	for (i = sglob->last_can_filter; i; i = i->next)
		if ((rx->id & i->mask) == i->id)
			send_can2client(rx, i->client->fd, i->flags);
}

inline int serve(void)
{
	int       i, n, maxfd, maxi, initfd, client_fd, byte_gelesen;
	char      buf[MAX_ZEICHEN];
	char      msgbuf[MAX_ZEICHEN];
	pid_t     pid;
	fd_set    rmenge, allmenge;
        canmsg_t  rx;

	FD_ZERO(&allmenge);
	if ( (initfd = serv_init(CS_NAME)) < 0)
		syslog(LOG_ERR, "can't get socket");
	FD_SET(initfd, &allmenge);
	FD_SET(sglob->can_fd, &allmenge);
	maxfd = initfd;
	maxi  = -1;

	while (1) {
		rmenge = allmenge;
		if ( (n = select(maxfd+1, &rmenge, NULL, NULL, NULL)) < 0)
			syslog(LOG_ERR, "select error");

		if (FD_ISSET(initfd, &rmenge)) {  /* Neue Client-request zulassen */
			if ( (client_fd = serv_accept(initfd, &pid)) < 0) {
				syslog(LOG_ERR, "connection refused - fd: %d", client_fd);
				continue;
			}
			i = client_add(client_fd, pid);
			FD_SET(client_fd, &allmenge);
			if (client_fd > maxfd)
				maxfd =client_fd;  /* groesster Filedeskr. fuer select */
			if (i > maxi)
				maxi = i;    /* Neue Anzahl von Clients im Array client[] */
			syslog(LOG_DEBUG, "new connection: pid %d, fd %d", pid, client_fd);
			continue;
		} else if (FD_ISSET(sglob->can_fd, &rmenge)) {
	
			can_Receive(sglob->can_fd, &rx);
			can_dispatch(&rx);
			//look thro filter queue
			//id,mask,client_fd
			//send msg - format?
			can_fmt_msg(msgbuf,&rx);
			syslog(LOG_ERR,msgbuf);

		} else
		
		for (i=0; i<=maxi; i++) {  /* Durchlaufen des client[]-Arrays */
			if ( (client_fd = sglob->client[i].fd) >= 0 &&
					FD_ISSET(client_fd, &rmenge)) {
				if ( (byte_gelesen = read(client_fd, buf, MAX_ZEICHEN)) < 0)
					syslog(LOG_ERR, "read error fd %d", client_fd);
				else if (byte_gelesen == 0) {
					syslog(LOG_DEBUG, "connection closed: pid %d, fd %d",
							sglob->client[i].pid, client_fd);
					client_del(&sglob->client[i]);
					FD_CLR(client_fd, &allmenge);
					close(client_fd);
				} else
					request(buf, byte_gelesen, &sglob->client[i]);
			}
		}
	}
}

/*----- request ---------------------------------------------------*/
/* client can:
 * send a can-message 
 * set a filter
 */

static void  request(char *buf, int byte_gelesen, Client *client)
{
	char     *argv[MAX_ARGC];
	int      argc;
	char     *token;
	int      client_fd  = client->fd;
	canmsg_t tx;

	if (buf[byte_gelesen-1] != '\0') {
		sprintf(err_msg, "request von %d ohne abschl. \\0: %*.*s\n",
				client->pid, byte_gelesen, byte_gelesen, buf);
		send_err(client_fd, -1, err_msg);
		return;
	}
	syslog(LOG_DEBUG, "request: %s, von pid %d", buf, client->pid);
	if (buf_argv(buf, &argc, argv) < 0) {
		send_err(client_fd, -1, err_msg);
		syslog(LOG_DEBUG, err_msg);
		return;
	}
	// parse args - adopted from CANterm 
	// (c) 1998 by C.Schroeter (clausi@chemie.fu-berlin.de)
	// 
	token = argv[0];
	if (token[0] == 'f' && !strncmp(argv[0],"filter",6)){
		struct can_filter *f = filter_create(sglob->last_can_filter);
		f->id = addrtoi(argv[1]);
		f->mask = addrtoi(argv[2]);
		f->client = client;
		sscanf(argv[3],"%d", &(f->flags)); 
		sglob->last_can_filter = f;
		sprintf(err_msg, "filter: %s\n", argv[1]);
		send_ok(client_fd, err_msg);
		syslog(LOG_DEBUG, err_msg);
	}
	else if( token[0] == 'R' && !strncmp(token,"RTR",3) ) {
		int sent=0;

		tx.flags= MSG_RTR;
		tx.id = addrtoi(argv[1]);
		sent = can_Send(sglob->can_fd,&tx);
		sprintf(err_msg, "RTR: %s\n", argv[1]);
		send_ok(client_fd, err_msg);
		syslog(LOG_DEBUG, err_msg);
	}

	//syslog(LOG_DEBUG, "Filesdeskr. %d fuer %s ueber Filesdeskr. %d geschickt",
	//		neufd, argv[1], client_fd);
}

static int  buf_argv(char *buf, int *argc, char *argv[])
{
	char    *zgr;

	if (strtok(buf, " \t\n") == NULL)
		return(-1);

	argv[*argc=0] = buf;
	while ( (zgr = strtok(NULL, " \t\n")) != NULL) {
		if (++*argc >= MAX_ARGC-1)
			return(-1);
		argv[*argc] = zgr;
	}
	argv[++*argc] = NULL;
	return(0);
}
