#include   <sys/types.h>
#include   <sys/socket.h>
#include   <sys/stat.h>
#include   <sys/un.h>
#include   <stddef.h>
#include   <time.h>
#include "caracad.h"

int  serv_init(const char *name)
{
	int                  fd, len;
	struct sockaddr_un   unix_adr;

	if ( (fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
		return(-1);

	unlink(name);

	memset(&unix_adr, 0, sizeof(unix_adr));
	unix_adr.sun_family = AF_UNIX;
	strcpy(unix_adr.sun_path, name);
	len = strlen(unix_adr.sun_path) + sizeof(unix_adr.sun_family);

	if (bind(fd, (struct sockaddr *) &unix_adr, len) < 0)
		return(-2);

	if (listen(fd, 5) < 0)
		return(-3);

	return(fd);
}

int  serv_accept(int initfd, pid_t *pidzgr)
{
	int                  client_fd, len;
	time_t               ablauf_zeit;
	struct sockaddr_un   unix_adr;
	struct stat          statpuff;

	len = sizeof(unix_adr);
	if ((client_fd = accept(initfd, (struct sockaddr *) &unix_adr, &len)) < 0)
		return(-1);

	len -= sizeof(unix_adr.sun_family);
	unix_adr.sun_path[len] = '\0';

	if (stat(unix_adr.sun_path, &statpuff) < 0)
		return(-2);
        /*
	if ((statpuff.st_mode & (S_IRWXG | S_IRWXO)) ||
			(statpuff.st_mode & S_IRWXU) != S_IRWXU)
		return(-3);    */

	/*--- Name vom Client darf nicht aelter als 30 Sek. sein ---*/
	ablauf_zeit = time(NULL) - 30;
	if (statpuff.st_atime < ablauf_zeit ||
			statpuff.st_ctime < ablauf_zeit ||
			statpuff.st_mtime < ablauf_zeit)
		return(-4);   /* i-node ist zu alt */

	if (pidzgr != NULL)
		sscanf(unix_adr.sun_path,"/var/tmp/%d",pidzgr);

	unlink(unix_adr.sun_path);

	return(client_fd);
}

int  cli_connect(const char *name)
{
	int                  fd, len;
	struct sockaddr_un   unix_adr;

	/*--- Kreieren eines Unix domain stream socket ---*/
	if ( (fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
		return(-1);

	/*--- Schreiben der socket adress Struktur mit eigener Adresse ---*/
	memset(&unix_adr, 0, sizeof(unix_adr));
	unix_adr.sun_family = AF_UNIX;
	sprintf(unix_adr.sun_path, "/var/tmp/%05d", getpid());

	len = strlen(unix_adr.sun_path) + sizeof(unix_adr.sun_family);
	if (len != 16)
		return(-1);
	//fehler_meld(FATAL_SYS, "ungueltige Laenge");

	unlink(unix_adr.sun_path);

	/*--- Zuordnen des Namens zum Filedeskriptor ---*/
	if (bind(fd, (struct sockaddr *) &unix_adr, len) < 0)
		return(-2);

	if (chmod(unix_adr.sun_path, S_IRWXU) < 0)
		return(-3);

	/*--- Schreiben der socket adress Struktur mit Server-Adresse ---*/
	memset(&unix_adr, 0, sizeof(unix_adr));
	unix_adr.sun_family = AF_UNIX;
	strcpy(unix_adr.sun_path, name);

	len = strlen(unix_adr.sun_path) + sizeof(unix_adr.sun_family);

	if (connect(fd, (struct sockaddr *) &unix_adr, len) < 0)
		return(-4);

	return(fd);
}
