#include <sys/socket.h>   /* struct msghdr */
#include <sys/uio.h>      /* struct iovec  */
#include <stddef.h>
#include "caracad.h"

/* Groesse des Kontrollpuffers zum Senden/Empfangen eines Filedeskr. */
#define KONTROLLAENGE  (sizeof(struct cmsghdr) + sizeof(int))

static struct cmsghdr  *cmzgr = NULL;  /* beim erstemal malloc hierfuer */

int addrtoi(char *addr)
{
	int id,f,dev;
	
	sscanf(addr,"%d.%d.%d",&f,&id,&dev);
	return (f << 8) + (id << 3) + dev;
}

int can_fmt_msg(char *res, canmsg_t *rx )
{
	char *buf = res;
	int i;
	
	buf += sprintf(buf,"%lx.%lx.%lx:",(rx->id>>8)&0x7 , (rx->id >>3) & 0x1f,rx->id & 0x7);
	for (i=0;i<rx->length;i++) buf += sprintf(buf,"%d,", rx->data[i]);
	*(buf-1) = '\0';
	return buf-res;
}

int send_can2client(canmsg_t *rx,int fd,int format)
{
	char msgbuf[50];
	write(fd,msgbuf,can_fmt_msg(msgbuf,rx));	
}

/*----- send_fd --------------------------------------------------------
 *         sendet einen Filedeskriptor an anderen Prozess
 *         wenn fd<0, so wird -fd als Fehler-Status geschickt         */
int  send_fd(int spipefd , int fd)
{
	struct iovec    iov[1];
	struct msghdr   message;
	char            protokoll[2] = { 0, 0 };

	iov[0].iov_base = protokoll;
	iov[0].iov_len  = 2;
	message.msg_iov     = iov;
	message.msg_iovlen  = 1;
	message.msg_name    = NULL;
	message.msg_namelen = 0;

	if (fd < 0) {
		message.msg_control    = NULL;
		message.msg_controllen = 0;
		protokoll[1] = -fd;  /* Status != 0 bedeutet Fehler */
		if (protokoll[1] == 0)
			protokoll[1] = 1;  /* Abfangen von Ueberlaeufen */
	} else {
		if (cmzgr == NULL && (cmzgr = malloc(KONTROLLAENGE)) == NULL)
			return(-1);
		cmzgr->cmsg_level  = SOL_SOCKET;
		cmzgr->cmsg_type   = SCM_RIGHTS;
		cmzgr->cmsg_len    = KONTROLLAENGE;
		message.msg_control    = (caddr_t) cmzgr;
		message.msg_controllen = KONTROLLAENGE;
		*(int *)CMSG_DATA(cmzgr) = fd;  /* zu schickender Filedeskriptor */
	}

	if (sendmsg(spipefd, &message, 0) != 2)
		return(-1);

	return(0);
}

/*----- empfang_fd -----------------------------------------------------
 *         empfaengt einen Filedeskriptor von einem anderen Prozess.
 *         Zusaetzlich empfangene Daten werden von
 *           (*benutzerfunk)(STDERR_FILENO, puffer, byte_gelesen)
 *         verarbeitet.                                               */
int rec_fd(int spipefd, ssize_t (*benutzerfunk)(int, const void *, size_t))
{
	int                neufd, byte_gelesen, status=-1;
	char               *zgr, puffer[MAX_ZEICHEN];
	struct iovec       iov[1];
	struct msghdr      message;

	while (1) {
		iov[0].iov_base = puffer;
		iov[0].iov_len  = sizeof(puffer);
		message.msg_iov     = iov;
		message.msg_iovlen  = 1;
		message.msg_name    = NULL;
		message.msg_namelen = 0;
		if (cmzgr == NULL && (cmzgr = malloc(KONTROLLAENGE)) == NULL)
			return(-1);
		message.msg_control    = (caddr_t) cmzgr;
		message.msg_controllen = KONTROLLAENGE;

		if ( (byte_gelesen = recvmsg(spipefd, &message, 0)) < 0)
			syslog(LOG_ERR, "recvmsg-Fehler");
		else if (byte_gelesen == 0) {
			syslog(LOG_WARNING, "Verbindung abgebrochen (durch Server)");
			return(-1);
		}

		/* Durchlaufen des ganzen Puffers, wobei die eigentlichen Daten
		   mit einem Null-Byte abgeschlossen sind, dem dann der
		   Status folgt. Ein Status-Wert von 0 bedeutet dabei, dass
		   ein Filedeskriptor empfangen wird.                          */
		for (zgr=puffer; zgr < &puffer[byte_gelesen]; ) {
			if (*zgr++ == 0) {
				if (zgr != &puffer[byte_gelesen-1])
					syslog(LOG_WARNING, "message inkonsistent");
				status = *zgr & 0xff;
				if (status == 0) {
					if (message.msg_controllen != KONTROLLAENGE)
						syslog(LOG_WARNING, "message inkonsistent");
					neufd = *(int *)CMSG_DATA(cmzgr);
				} else
					neufd = -status;
				byte_gelesen -= 2;
			}
		}
		if (byte_gelesen > 0)
			if ( (*benutzerfunk)(STDERR_FILENO, puffer, byte_gelesen)
					!= byte_gelesen)
				return(-1);
		if (status >= 0)
			return(neufd);
	}
}

/*----- send_fehl ------------------------------------------------------
 *         sendet mittels dem beschriebenen Protokoll einen Fehler.
 *         Diese Routine wird benutzt, wenn beabsichtigt war, einen
 *         Filedeskriptor zu schicken, aber ein Fehler aufgetreten ist.*/
int send_err(int spipefd, int status, const char *errmsg)
{
	int     n;

	if ( (n=strlen(errmsg)) > 0)
		if (write(spipefd, errmsg, n) != n) /* Senden der Fehlermeldung */
			return(-1);

	if (status >= 0)
		status = -1;  /* Status muss negativ sein */

	if (send_fd(spipefd, status) < 0)
		return(-1);

	return(0);
}

int send_ok(int spipefd, const char *errmsg)
{
	struct iovec    iov[2];
	char            ok[3] = { 'O', 'K', ' ' };

	iov[0].iov_base = ok;
	iov[0].iov_len  = 3;
	iov[1].iov_base = errmsg;
	iov[1].iov_len  = strlen(errmsg);

	if (writev(spipefd, iov, 2) != iov[1].iov_len+3) 
		return(-1);
	return(0);
}
